"""Generic data types for information retrieval"""

from abc import ABC, abstractmethod
from functools import cached_property
import logging
from pathlib import Path
from attrs import define
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Type, TYPE_CHECKING
import random
from experimaestro import Config
from datamaestro.definitions import datatasks, Param, Meta
from datamaestro.data import Base
from datamaestro_text.utils.files import auto_open
from datamaestro_text.utils.iter import BatchIterator
from datamaestro.record import record_type, RecordType
from .base import (  # noqa: F401
    # Record items
    IDItem,
    TextItem,
    InternalIDItem,
    TopicRecord,
    DocumentRecord,
    SimpleTextItem,
    ScoredItem,
    # Create records
    create_record,
    # Other things
    AdhocAssessment,
    AdhocAssessedTopic,
)

#: A adhoc run dictionary (query id -> doc id -> score)
AdhocRunDict = dict[str, dict[str, float]]


class Documents(Base):
    """A set of documents with identifiers

    See `IR Datasets <https://ir-datasets.com/index.html>`_ for the list of query classes
    """

    count: Meta[Optional[int]]
    """Number of documents"""

    def iter(self) -> Iterator[DocumentRecord]:
        """Returns an iterator over documents"""
        raise self.iter_documents()

    def iter_documents(self) -> Iterator[DocumentRecord]:
        return self.iter()

    def iter_documents_from(self, start=0) -> Iterator[DocumentRecord]:
        """Iterate over a range of documents

        Can be specialized in a subclass for faster access

        :param start: The starting document, defaults to 0
        :return: An iterator
        """
        iter = self.iter()
        if start > 0:
            logging.info("skipping %d documents", start + 1)
            for _ in range(start + 1):
                next(iter)

        return iter

    def iter_ids(self) -> Iterator[str]:
        """Iterates over document ids

        By default, use iter_documents, which is not really efficient.
        """
        for doc in self.iter():
            yield doc[IDItem].id

    @property
    def documentcount(self):
        """Returns the number of terms in the index"""
        if self.count is not None:
            return self.count

        raise NotImplementedError(f"For class {self.__class__}")

    @property
    @abstractmethod
    def document_recordtype(self) -> Type[DocumentRecord]:
        """The class for documents"""
        ...


class DocumentStore(Documents):
    """A document store

    A document store can
    - match external/internal ID
    - return the document content
    - return the number of documents
    """

    def docid_internal2external(self, docid: int):
        """Converts an internal collection ID (integer) to an external ID"""
        raise NotImplementedError(f"For class {self.__class__}")

    def document_int(self, internal_docid: int) -> DocumentRecord:
        """Returns a document given its internal ID"""
        docid = self.docid_internal2external(internal_docid)
        return self.document_ext(docid)

    def document_ext(self, docid: str) -> DocumentRecord:
        """Returns a document given its external ID"""
        raise NotImplementedError(f"document() in {self.__class__}")

    def documents_ext(self, docids: List[str]) -> List[DocumentRecord]:
        """Returns documents given their external ID

        By default, just look using `document_ext`, but some store might
        optimize batch retrieval
        """
        return [self.document_ext(docid) for docid in docids]

    def iter_sample(
        self, randint: Optional[Callable[[int], int]]
    ) -> Iterator[DocumentRecord]:
        """Sample documents from the dataset"""
        length = self.documentcount
        randint = randint or (lambda max: random.randint(0, max - 1))
        while True:
            yield self.document_int(randint(length))


class AdhocIndex(DocumentStore):
    """An index can be used to retrieve documents based on terms"""

    @property
    def termcount(self):
        """Returns the number of terms in the index"""
        raise NotImplementedError(f"For class {self.__class__}")

    def term_df(self, term: str):
        """Returns the document frequency"""
        raise NotImplementedError(f"For class {self.__class__}")


class Topics(Base, ABC):
    """A set of topics with associated IDs"""

    @abstractmethod
    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        ...

    def __iter__(self):
        return self.iter()

    def count(self) -> Optional[int]:
        """Returns the number of topics if known"""
        return None

    @property
    @abstractmethod
    def topic_recordtype(self) -> Type[TopicRecord]:
        """The class for topics"""


AdhocTopics = Topics


class TopicsStore(Topics):
    """Adhoc topics store"""

    @abstractmethod
    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""

    @abstractmethod
    def topic_ext(self, external_topic_id: int) -> TopicRecord:
        """Returns a document given its external ID"""


class AdhocAssessments(Base, ABC):
    """Ad-hoc assessments (qrels)"""

    def iter(self) -> Iterator[AdhocAssessedTopic]:
        """Returns an iterator over assessments"""
        raise NotImplementedError(f"For class {self.__class__}")


class AdhocRun(Base):
    """IR adhoc run"""

    @abstractmethod
    def get_dict(self) -> "AdhocRunDict":
        """Get the run as a dictionary query ID -> doc ID -> score"""
        ...


class AdhocResults(Base):
    def get_results(self) -> Dict[str, float]:
        """Returns the aggregated results

        :return: Returns a dictionary where each metric (keys) is associated
            with a value
        """
        raise NotImplementedError(f"For class {self.__class__}")


@datatasks("information retrieval")
class Adhoc(Base):
    """An Adhoc IR collection with documents, topics and their assessments"""

    documents: Param[Documents]
    """The set of documents"""

    topics: Param[Topics]
    """The set of topics"""

    assessments: Param[AdhocAssessments]
    """The set of assessments (for each topic)"""


class RerankAdhoc(Adhoc):
    """Re-ranking ad-hoc task based on an existing run"""

    run: Param[AdhocRun]
    """The run to re-rank"""


class Measure(Config):
    """An Information Retrieval measure"""

    pass


Triplets = Tuple[TopicRecord, DocumentRecord, DocumentRecord]


class TrainingTriplets(Base, ABC):
    """Triplet for training IR systems: query / query ID, positive document,
    negative document"""

    def iter(self) -> Iterator[Triplets]:
        """Returns an iterator over (topic, document 1, document) triplets"""
        raise NotImplementedError(f"For class {self.__class__}")

    def batch_iter(self, size: int) -> Iterator[List[Triplets]]:
        """Returns an iterator over batches of triplets

        The default implementation just concatenates triplets using `iter`, but
        some classes might use more efficient ways to provide batches of data
        """
        return BatchIterator(self.iter(), size)

    def count(self):
        """Returns the number of triplets or None"""
        return None

    @property
    @abstractmethod
    def topic_recordtype(self) -> RecordType:
        """The set of records for topics"""
        ...

    @property
    @abstractmethod
    def document_recordtype(self) -> RecordType:
        """The class for documents"""
        ...


class TrainingTripletsLines(TrainingTriplets):
    """Training triplets with one line per triple (query texts)"""

    sep: Meta[str]
    path: Param[Path]

    doc_ids: Meta[bool]
    """True if we have documents IDs"""

    topic_ids: Meta[bool]
    """True if we have query IDs"""

    def iter(self) -> Iterator[Triplets]:
        with auto_open(self.path, "rt") as fp:
            for line in fp:
                q, pos, neg = line.strip().split(self.sep)
                yield self._topic(q), self._doc(pos), self._doc(neg)

    @cached_property
    def _doc(self):
        return lambda doc: self.document_recordtype(
            IDItem(doc) if self.doc_ids else SimpleTextItem(doc)
        )

    @cached_property
    def _topic(self):
        return lambda q: self.topic_recordtype(
            IDItem(q) if self.topic_ids else SimpleTextItem(q)
        )

    @cached_property
    def topic_recordtype(self) -> Type[TopicRecord]:
        """The class for topics"""
        return record_type(IDItem) if self.topic_ids else record_type(SimpleTextItem)

    @cached_property
    def document_recordtype(self) -> Type[DocumentRecord]:
        """The class for documents"""
        return record_type(IDItem) if self.doc_ids else record_type(SimpleTextItem)


@define(kw_only=True)
class PairwiseSample(ABC):
    """A a query with positive and negative samples"""

    topics: List[TopicRecord]
    """The topic(s)"""

    positives: List[DocumentRecord]
    """Relevant documents"""

    negatives: Dict[str, List[DocumentRecord]]
    """Non relevant documents, organized in a dictionary where keys
    are the algorithm used to retrieve the negatives"""


class PairwiseSampleDataset(Base, ABC):
    """Datasets where each record is a query with positive and negative samples"""

    @abstractmethod
    def iter(self) -> Iterator[PairwiseSample]:
        ...



from pathlib import Path
from typing import Iterator, Optional
import re
from datamaestro_text.data.ir import AdhocRunDict
from datamaestro_text.data.ir.base import (
    AdhocAssessedTopic,
    TopicRecord,
    SimpleAdhocAssessment,
    IDItem,
)
from datamaestro_text.data.ir.formats import TrecTopicRecord, TrecTopic

# --- Runs


def parse_run(path: Path) -> AdhocRunDict:
    results = {}
    with path.open("rt") as f:
        for line in f:
            query_id, _q0, doc_id, _rank, score, _model_id = re.split(
                r"\s+", line.strip()
            )
            results.setdefault(query_id, {})[doc_id] = score

    return results


def write_run_dict(run: AdhocRunDict, run_path: Path):
    """Write run dict"""
    with run_path.open("wt") as f:
        for query_id, scored_documents in run.items():
            scored_documents = list(
                [(doc_id, score) for doc_id, score in scored_documents.items()]
            )
            scored_documents.sort(key=lambda x: x[1], reverse=True)
            for ix, (doc_id, score) in enumerate(scored_documents):
                f.write(f"{query_id} Q0 {doc_id} {ix + 1} {score} run\n")


# --- Assessments


def parse_qrels(path: Path) -> Iterator[AdhocAssessedTopic]:
    with path.open("rt") as fp:
        _qid = None
        assessments = []

        for line in fp:
            qid, _, docno, rel = re.split(r"\s+", line.strip())
            if qid != _qid:
                if _qid is not None:
                    yield AdhocAssessedTopic(_qid, assessments)
                _qid = qid
                assessments = []
            assessments.append(SimpleAdhocAssessment(docno, int(rel)))

        yield AdhocAssessedTopic(_qid, assessments)


# ---- TOPICS


def cleanup(s: Optional[str]) -> str:
    return s.replace("\t", " ").strip() if s is not None else ""


def parse_query_format(file, xml_prefix=None) -> Iterator[TopicRecord]:
    """Parse TREC XML query format"""
    if xml_prefix is None:
        xml_prefix = ""

    if hasattr(file, "read"):
        num, title, desc, narr, reading = None, None, None, None, None
        for line in file:
            if line.startswith("**"):
                # translation comment in older formats (e.g., TREC 3 Spanish track)
                continue
            elif line.startswith("</top>"):
                if num:
                    yield TrecTopicRecord(
                        IDItem(num),
                        TrecTopic(cleanup(title), cleanup(desc), cleanup(narr)),
                    )
                num, title, desc, narr, reading = None, None, None, None, None
            elif line.startswith("<num>"):
                num = line[len("<num>") :].replace("Number:", "").strip()
                reading = None
            elif line.startswith(f"<{xml_prefix}title>"):
                title = line[len(f"<{xml_prefix}title>") :].strip()
                if title == "":
                    reading = "title"
                else:
                    reading = None
            elif line.startswith(f"<{xml_prefix}desc>"):
                desc = ""
                reading = "desc"
            elif line.startswith(f"<{xml_prefix}narr>"):
                narr = ""
                reading = "narr"
            elif reading == "desc":
                desc += line.strip() + " "
            elif reading == "narr":
                narr += line.strip() + " "
            elif reading == "title":
                title += line.strip() + " "
    else:
        with open(file, "rt") as f:
            yield from parse_query_format(f)



from abc import ABC, abstractmethod
from enum import Enum
from datamaestro_text.data.ir.base import IDItem, SimpleTextItem
from experimaestro import Param
from typing import Dict, Generic, Iterator, List, Optional, Sequence, Tuple
from attr import define
from datamaestro.record import record_type
from datamaestro.data import Base
from datamaestro.record import Record, Item
from datamaestro_text.data.ir import TopicRecord, Topics
from datamaestro_text.utils.iter import FactoryIterable, LazyList, RangeView

# ---- Basic types


class EntryType(Item, Enum):
    """Type of record"""

    USER_QUERY = 0
    SYSTEM_ANSWER = 1
    CLARIFYING_QUESTION = 2


class DecontextualizedItem(Item):
    """A topic record with decontextualized versions of the topic"""

    @abstractmethod
    def get_decontextualized_query(self, mode=None) -> str:
        """Returns the decontextualized query"""
        ...


@define
class SimpleDecontextualizedItem(DecontextualizedItem):
    """A topic record with one decontextualized version of the topic"""

    decontextualized_query: str

    @abstractmethod
    def get_decontextualized_query(self, mode=None) -> str:
        """Returns the decontextualized query"""
        assert mode is None

        return self.decontextualized_query


@define
class DecontextualizedDictItem(DecontextualizedItem):
    """A conversation entry providing decontextualized version of the user query"""

    default_decontextualized_key: str

    decontextualized_queries: Dict[str, str]

    def get_decontextualized_query(self, mode=None):
        return self.decontextualized_queries[mode or self.default_decontextualized_key]


@define
class AnswerEntry(Item):
    """A system answer"""

    answer: str
    """The system answer"""


@define
class AnswerDocumentID(Item):
    """An answer as a document ID"""

    document_id: str


@define
class AnswerDocumentURL(Item):
    """An answer as a document ID"""

    url: str


@define
class RetrievedEntry(Item):
    """List of system-retrieved documents and their relevance"""

    documents: List[str]
    """List of retrieved documents"""

    relevant_documents: Optional[Dict[int, Tuple[Optional[int], Optional[int]]]] = None
    """List of relevance status (optional), with start/stop position"""


@define
class ClarifyingQuestionEntry(Item):
    """A system-generated clarifying question"""

    pass


#: The conversation
ConversationHistory = Sequence[Record]


@define
class ConversationHistoryItem(Item):
    """A user interaction contextualized within a conversation"""

    history: ConversationHistory
    """The history"""


# ---- Abstract conversation representation


class ConversationNode:
    @abstractmethod
    def entry(self) -> Record:
        """The current conversation entry"""
        ...

    @abstractmethod
    def history(self) -> ConversationHistory:
        """Preceding conversation entries, from most recent to more ancient"""
        ...

    @abstractmethod
    def parent(self) -> Optional["ConversationNode"]: ...

    @abstractmethod
    def children(self) -> List["ConversationNode"]: ...


class ConversationTree(ABC):
    """Represents a conversation tree"""

    @abstractmethod
    def root(self) -> ConversationNode: ...

    @abstractmethod
    def __iter__(self) -> Iterator[ConversationNode]:
        """Iterates over conversation nodes"""
        ...


# ---- A conversation tree


class SingleConversationTree(ConversationTree, ABC):
    """Simple conversations, based on a sequence of entries"""

    id: str
    history: List[Record]

    def __init__(self, id: Optional[str], history: List[Record]):
        """Create a simple conversation

        :param history: The entries, in **reverse** order (i.e. more ancient first)
        """
        self.history = history or []
        self.id = id

    def add(self, entry: Record):
        self.history.insert(0, entry)

    def __iter__(self) -> Iterator[ConversationNode]:
        """Iterates over the conversation (starting with the beginning)"""
        for ix in reversed(range(len(self.history))):
            yield SingleConversationTreeNode(self, ix)

    def root(self):
        return SingleConversationTreeNode(self, len(self.history) - 1)


@define
class SingleConversationTreeNode(ConversationNode):
    tree: SingleConversationTree
    index: int

    @property
    def entry(self) -> Record:
        return self.tree.history[self.index]

    @entry.setter
    def entry(self, record: Record):
        try:
            self.tree.history[self.index] = record
        except Exception as e:
            print(e)
            raise

    def history(self) -> Sequence[Record]:
        return self.tree.history[self.index + 1 :]

    def parent(self) -> Optional[ConversationNode]:
        return (
            SingleConversationTreeNode(self.tree, self.index + 1)
            if self.index < len(self.tree.history) - 1
            else None
        )

    def children(self) -> List[ConversationNode]:
        return (
            [SingleConversationTreeNode(self.tree, self.index - 1)]
            if self.index > 0
            else []
        )


class ConversationTreeNode(ConversationNode, ConversationTree):
    """A conversation tree node"""

    entry: Record
    _parent: Optional["ConversationTreeNode"]
    _children: List["ConversationTreeNode"]

    def __init__(self, entry):
        self.entry = entry
        self._parent = None
        self._children = []

    def add(self, node: "ConversationTreeNode") -> "ConversationTreeNode":
        self._children.append(node)
        node._parent = self
        return node

    def conversation(self, skip_self: bool) -> ConversationHistory:
        def iterator():
            current = self.parent() if skip_self else self
            while current is not None:
                yield current.entry
                current = current.parent()

        return LazyList(FactoryIterable(iterator))

    def __iter__(self) -> Iterator["ConversationTreeNode"]:
        """Iterates over all conversation tree nodes (pre-order)"""
        yield self.entry
        for child in self._children:
            yield from child

    def parent(self) -> Optional[ConversationNode]:
        return self._parent

    def children(self) -> List[ConversationNode]:
        return self._children

    def root(self):
        return self


class ConversationDataset(Base, ABC):
    """A dataset made of conversations"""

    @abstractmethod
    def __iter__(self) -> Iterator[ConversationTree]:
        """Return an iterator over conversations"""
        ...


class ConversationUserTopics(Topics):
    """Extract user topics from conversations"""

    conversations: Param[ConversationDataset]

    topic_recordtype = record_type(IDItem, SimpleTextItem)

    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        # Extracts topics from conversations, Each user query is a topic (can perform retrieval on it)
        # TODO: merge with xpmir.learning.DatasetConversationBase -> same logic
        
        records: List[TopicRecord] = []
        for conversation in self.conversations.__iter__():
            nodes = [
                node
                for node in conversation
                if node.entry[EntryType] == EntryType.USER_QUERY
            ]
            for node in nodes:
                records.append(
                    node.entry.update(ConversationHistoryItem(node.history()))
                )
        return iter(records)


import logging
from typing import Optional, Type, Callable, Iterator
from ir_datasets.indices import PickleLz4FullStore
from datamaestro.download import Resource
from datamaestro.utils import FileChecker
from pathlib import Path
import urllib3


class lz4docstore_downloader(Resource):
    """Uses ir_datasets Lz4FullStore to build a document store for a stream of documents"""

    def __init__(
        self,
        varname: str,
        url: str,
        iter_factory: Callable[[Path], Iterator],
        doc_cls: Type,
        lookup_field: str,
        *,
        count_hint: Optional[int] = None,
        size: Optional[int] = None,
        checker: FileChecker = None,
    ):
        super().__init__(varname)
        self.iter_factory = iter_factory
        self.url = url
        self.doc_cls = doc_cls
        self.size = size
        self.lookup_field = lookup_field
        self.count_hint = count_hint
        self.checker = checker

        p = urllib3.util.parse_url(self.url)
        assert p is not None
        self.name = Path(p.path).with_suffix("").name

    def prepare(self):
        return self.definition.datapath / self.name

    def download(self, force=False):
        # Creates directory if needed
        destination = self.definition.datapath / self.name
        destination.mkdir(exist_ok=True)

        # Early exit
        if (destination / "done").is_file() and not force:
            return True

        # Download (cache)
        logging.info("Building the document index")
        with self.context.downloadURL(self.url, size=self.size) as file:
            # Checks the file
            if self.checker:
                self.checker.check(file.path)

            # Builds the LZ4 store
            store = PickleLz4FullStore(
                destination,
                lambda: self.iter_factory(Path(file.path)),
                self.doc_cls,
                lookup_field=self.lookup_field,
                index_fields=[self.lookup_field],
                key_field_prefix=None,
                size_hint=None,
                count_hint=self.count_hint,
            )
            store.build()

            # All good!
            (destination / "done").touch()


class lz4docstore_builder(Resource):
    def __init__(
        self,
        name: str,
        iter_factory: Callable[[], Iterator],
        doc_cls: Type,
        lookup_field: str,
        *,
        count_hint: Optional[int] = None,
    ):
        """Uses ir_datasets Lz4FullStore to build a document store for a stream of documents

        :param name: The name of the variable for path construction
        :param iter_factory: Iterator over documents
        :param doc_cls: The class of documents (must be a dataclass because of how ir-datasets works)
        :param lookup_field: Which field to use for lookup
        :param count_hint: Number of documents (hint), defaults to None
        """
        super().__init__(name)
        self.iter_factory = iter_factory
        self.doc_cls = doc_cls
        self.lookup_field = lookup_field
        self.count_hint = count_hint

    def prepare(self):
        return self.definition.datapath / self.varname

    def download(self, force=False):
        # Creates directory if needed
        destination = self.definition.datapath / self.varname
        destination.mkdir(exist_ok=True)

        # Early exit
        if (destination / "done").is_file() and not force:
            return True

        # Download (cache)
        logging.info("Building the document index")

        # Builds the LZ4 store
        store = PickleLz4FullStore(
            destination,
            lambda: self.iter_factory(),
            self.doc_cls,
            lookup_field=self.lookup_field,
            index_fields=[self.lookup_field],
            key_field_prefix=None,
            size_hint=None,
            count_hint=self.count_hint,
        )
        store.build()

        # All good!
        (destination / "done").touch()



from abc import ABC, abstractmethod
from attrs import define
from typing import List
from datamaestro.record import Record, Item, record_type


TopicRecord = DocumentRecord = Record


@define()
class ScoredItem(Item):
    """A score associated with the document"""

    score: float
    """A retrieval score associated with this record (e.g. of the first-stage
    retriever)"""


class TextItem(Item, ABC):
    @property
    @abstractmethod
    def text(self) -> str:
        """Returns the text"""


@define
class SimpleTextItem(TextItem):
    """A topic/document with a text record"""

    text: str


@define
class InternalIDItem(Item, ABC):
    """A topic/document with an internal ID"""

    id: int


@define
class IDItem(Item, ABC):
    """A topic/document with an external ID"""

    id: str


@define
class UrlItem(Item):
    """An url item"""

    url: str


@define
class AdhocAssessment:
    doc_id: str


@define
class SimpleAdhocAssessment(AdhocAssessment):
    rel: float
    """Relevance (> 0 if relevant)"""


@define
class AdhocAssessedTopic:
    topic_id: str
    """The topic ID"""

    assessments: List[AdhocAssessment]
    """List of assessments for this topic"""


def create_record(*items: Item, id: str = None, text: str = None) -> Record:
    """Easy creation of a text/id item"""
    extra_items = []
    if id is not None:
        extra_items.append(IDItem(id))
    if text is not None:
        extra_items.append(SimpleTextItem(text))
    return Record(*items, *extra_items)



import re
from typing import Dict, List, Optional
from experimaestro import documentation, Param, Meta
from pathlib import Path
from datamaestro_text.data.ir import (
    AdhocRunDict,
    Documents,
    Topics,
    AdhocAssessments,
    AdhocRun,
    AdhocResults,
    Measure,
)
from datamaestro_text.data.ir.formats import TrecTopicRecord


class TrecTopics(Topics):
    path: Meta[Path]
    parts: Meta[List[str]]

    @documentation
    def iter(self):
        """Iterate over TREC adhoc topics"""
        import datamaestro_text.interfaces.trec as trec

        yield from trec.parse_query_format(self.path)

    @property
    def topic_recordtype(self):
        return TrecTopicRecord


class TrecAdhocAssessments(AdhocAssessments):
    path: Meta[Path]

    def trecpath(self):
        return self.path

    @documentation
    def iter(self):
        """Iterate over TREC adhoc topics"""
        import datamaestro_text.interfaces.trec as trec

        yield from trec.parse_qrels(self.path)


class TrecAdhocRun(AdhocRun):
    path: Param[Path]

    def get_dict(self) -> AdhocRunDict:
        import datamaestro_text.interfaces.trec as trec

        return trec.parse_run(self.path)


class TrecAdhocResults(AdhocResults):
    """Adhoc results (TREC format)"""

    metrics: Param[List[Measure]]
    """List of reported metrics"""

    results: Param[Path]
    """Main results"""

    detailed: Param[Optional[Path]]
    """Results per topic (if any)"""

    def get_results(self) -> Dict[str, float]:
        """Returns the results as a dictionary {metric_name: value}"""
        re_spaces = re.compile(r"\s+")

        results = {}
        with self.results.open("rt") as fp:
            for line in fp.readlines():
                metric, _, value = re_spaces.split(line.strip())
                results[metric] = value

        return results


class TipsterCollection(Documents):
    path: Param[Path]

    def iter(self):
        raise NotImplementedError()



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Dict, Iterator, List, NamedTuple, Tuple, Type

import ir_datasets
import ir_datasets.datasets as _irds
from datamaestro.record import RecordType, record_type
from experimaestro import Config, Meta, Option, Param
from experimaestro.compat import cached_property
from ir_datasets.formats import (
    GenericDoc,
    GenericDocPair,
    GenericQuery,
    TrecParsedDoc,
    TrecQuery,
)
from ir_datasets.indices import PickleLz4FullStore

import datamaestro_text.data.ir as ir
import datamaestro_text.data.ir.formats as formats
from datamaestro_text.data.conversation.base import (
    AnswerDocumentID,
    AnswerEntry,
    ConversationHistoryItem,
    ConversationTreeNode,
    DecontextualizedDictItem,
    EntryType,
)
from datamaestro_text.data.ir.base import (
    AdhocAssessedTopic,
    DocumentRecord,
    IDItem,
    Record,
    SimpleAdhocAssessment,
    SimpleTextItem,
    TopicRecord,
    UrlItem,
    create_record,
)

# Interface between ir_datasets and datamaestro:
# provides adapted data types


class IRDSId(Config):
    irds: Option[str]
    """The id to load the dataset from ir_datasets"""

    @classmethod
    def __xpmid__(cls):
        return f"ir_datasets.{cls.__qualname__}"

    @cached_property
    def dataset(self):
        return ir_datasets.load(self.irds)

    def iter(self) -> Iterator[Record]:
        """Returns an iterator over topics"""
        for query in self.dataset.queries_iter():
            yield self.factory(query)

    def count(self):
        return self.dataset.queries_count()


class AdhocAssessments(ir.AdhocAssessments, IRDSId):
    def iter(self):
        """Returns an iterator over assessments"""
        ds = self.dataset

        class Qrels(dict):
            def __missing__(self, key):
                qqrel = AdhocAssessedTopic(key, [])
                self[key] = qqrel
                return qqrel

        qrels = Qrels()
        for qrel in ds.qrels_iter():
            qrels[qrel.query_id].assessments.append(
                SimpleAdhocAssessment(qrel.doc_id, qrel.relevance)
            )

        return qrels.values()


class tuple_constructor:
    def __init__(self, target_cls: Type, id_field: str, *fields: str):
        self.id_field = id_field
        self.target_cls = target_cls
        self.fields = fields

    def check(self, source_cls: Type):
        source_fields = tuple(f for f in source_cls._fields if f != self.id_field)
        assert source_fields == self.fields, (
            "Internal error: Fields do not match, "
            f"source({source_cls.__qualname__})={source_fields}/{self.id_field} [vs] target={self.fields}"
        )

    def __call__(self, recordtype, entry):
        values = tuple(getattr(entry, key) for key in self.fields)
        return recordtype(
            IDItem(getattr(entry, self.id_field)), self.target_cls(*values)
        )


class Documents(ir.DocumentStore, IRDSId):
    CONVERTERS = {
        GenericDoc: tuple_constructor(SimpleTextItem, "doc_id", "text"),
        _irds.beir.BeirCordDoc: tuple_constructor(
            formats.CordDocument, "doc_id", "text", "title", "url", "pubmed_id"
        ),
        _irds.beir.BeirTitleDoc: tuple_constructor(
            formats.TitleDocument, "doc_id", "text", "title"
        ),
        _irds.beir.BeirTitleUrlDoc: tuple_constructor(
            formats.TitleUrlDocument, "doc_id", "text", "title", "url"
        ),
        _irds.beir.BeirToucheDoc: tuple_constructor(
            formats.Touche2020, "doc_id", "text", "title", "stance", "url"
        ),
        _irds.beir.BeirSciDoc: tuple_constructor(
            formats.SciDocs,
            "doc_id",
            "text",
            "title",
            "authors",
            "year",
            "cited_by",
            "references",
        ),
        _irds.msmarco_document.MsMarcoDocument: tuple_constructor(
            formats.MsMarcoDocument, "doc_id", "url", "title", "body"
        ),
        _irds.cord19.Cord19FullTextDoc: tuple_constructor(
            formats.CordFullTextDocument,
            "doc_id",
            "title",
            "doi",
            "date",
            "abstract",
            "body",
        ),
        _irds.nfcorpus.NfCorpusDoc: tuple_constructor(
            formats.NFCorpusDocument, "doc_id", "url", "title", "abstract"
        ),
        TrecParsedDoc: tuple_constructor(
            formats.TrecParsedDocument, "doc_id", "title", "body", "marked_up_doc"
        ),
        _irds.wapo.WapoDoc: tuple_constructor(
            formats.WapoDocument,
            "doc_id",
            "url",
            "title",
            "author",
            "published_date",
            "kicker",
            "body",
            "body_paras_html",
            "body_media",
        ),
        _irds.tweets2013_ia.TweetDoc: tuple_constructor(
            formats.TweetDoc,
            "doc_id",
            "text",
            "user_id",
            "created_at",
            "lang",
            "reply_doc_id",
            "retweet_doc_id",
            "source",
            "source_content_type",
        ),
        _irds.dpr_w100.DprW100Doc: tuple_constructor(
            formats.DprW100Doc,
            "doc_id",
            "text",
            "title",
        ),
        _irds.msmarco_passage_v2.MsMarcoV2Passage: tuple_constructor(
            formats.MsMarcoV2Passage,
            "doc_id",
            "text",
            "spans",
            "msmarco_document_id",
        ),
    }

    """Wraps an ir datasets collection -- and provide a default text
    value depending on the collection itself"""

    # List of fields
    # self.dataset.docs_cls()._fields

    def __getstate__(self):
        return (self.id, self.irds)

    def __setstate__(self, state):
        self.id, self.irds = state

    def iter(self) -> Iterator[ir.DocumentRecord]:
        """Returns an iterator over adhoc documents"""
        for doc in self.dataset.docs_iter():
            yield self.converter(self.document_recordtype, doc)

    def iter_documents_from(self, start=0):
        for doc in self.dataset.docs_iter()[start:]:
            yield self.converter(self.document_recordtype, doc)

    @property
    def documentcount(self):
        return self.dataset.docs_count()

    @cached_property
    def store(self):
        return self.dataset.docs_store()

    @cached_property
    def _docs(self):
        return self.dataset.docs_iter()

    def docid_internal2external(self, ix: int):
        return self._docs[ix].doc_id

    def document_ext(self, docid: str) -> DocumentRecord:
        return self.converter(self.document_recordtype, self.store.get(docid))

    def documents_ext(self, docids: List[str]) -> DocumentRecord:
        """Returns documents given their external IDs (optimized for batch)"""
        retrieved = self.store.get_many(docids)
        return [
            self.converter(self.document_recordtype, retrieved[docid])
            for docid in docids
        ]

    def document_int(self, ix):
        return self.converter(self.document_recordtype, self._docs[ix])

    @cached_property
    def document_recordtype(self):
        return record_type(IDItem, self.converter.target_cls)

    @cached_property
    def converter(self):
        converter = Documents.CONVERTERS[self.dataset.docs_cls()]
        converter.check(self.dataset.docs_cls())
        return converter


if hasattr(_irds, "miracl"):
    Documents.CONVERTERS[_irds.miracl.MiraclDoc] = tuple_constructor(
        formats.DocumentWithTitle, "doc_id", "text", "title"
    )


class LZ4DocumentStore(ir.DocumentStore, ABC):
    """A LZ4-based document store"""

    path: Param[Path]

    #: Lookup field
    lookup_field: Param[str]

    # Extra indexed fields (e.g. URLs)
    index_fields: List[str] = []

    @cached_property
    def store(self):
        return PickleLz4FullStore(
            self.path, None, self.data_cls, self.lookup_field, self.index_fields
        )

    @cached_property
    def _docs(self):
        return self.store.__iter__()

    def docid_internal2external(self, ix: int):
        return getattr(self._docs[ix], self.store._id_field)

    def document_ext(self, docid: str) -> DocumentRecord:
        return self.converter(self.store.get(docid))

    def documents_ext(self, docids: List[str]) -> DocumentRecord:
        """Returns documents given their external IDs (optimized for batch)"""
        retrieved = self.store.get_many(docids)
        return [self.converter(retrieved[docid]) for docid in docids]

    @abstractmethod
    def converter(self, data):
        """Converts a document from LZ4 tuples to a document record"""
        ...

    def iter(self) -> Iterator[DocumentRecord]:
        """Returns an iterator over documents"""
        return map(self.converter, self.store.__iter__())

    def iter_documents_from(self, start=0):
        return map(self.converter, self.store.__iter__()[start:])

    @cached_property
    def documentcount(self):
        if self.count:
            return self.count
        return self.store.count()


class SimpleJsonDocument(NamedTuple):
    id: str
    text: str


class LZ4JSONLDocumentStore(LZ4DocumentStore):
    jsonl_path: Meta[Path]
    """json-l based document store

    Each line is of the form
    ```json
    { "id": "...", "text": "..." }
    ```
    """

    def converter(self, data):
        return DocumentRecord(IDItem(data["id"]), SimpleTextItem(data["text"]))


class TopicsHandler(ABC):
    @abstractmethod
    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        ...

    @abstractmethod
    def topic_ext(self, external_topic_id: str) -> TopicRecord:
        """Returns a document given its external ID"""
        ...

    @abstractmethod
    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        ...


class SimpleTopicsHandler(TopicsHandler):
    def __init__(self, converter, topics: "Topics"):
        self.converter = converter
        self._topics = topics
        self.target_cls = converter.target_cls
        converter.check(topics.dataset.queries_cls())

    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        return self.topics_list[internal_topic_id]

    def topic_ext(self, external_topic_id: int) -> TopicRecord:
        """Returns a document given its external ID"""
        return self.topics_map[external_topic_id]

    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        yield from self.topics_list

    @cached_property
    def topics_map(self):
        return self.topics[0]

    @cached_property
    def topics_list(self):
        return self.topics[1]

    @cached_property
    def topics(self):
        topic_map = {}
        topic_list = []
        for query in self._topics.dataset.queries_iter():
            record = self.converter(self._topics.topic_recordtype, query)
            topic_map[query.query_id] = record
            topic_list.append(record)

        return topic_map, topic_list


class Topics(ir.TopicsStore, IRDSId):
    CONVERTERS = {
        GenericQuery: tuple_constructor(SimpleTextItem, "query_id", "text"),
        _irds.beir.BeirCovidQuery: tuple_constructor(
            formats.TrecTopic, "query_id", "text", "query", "narrative"
        ),
        _irds.beir.BeirUrlQuery: tuple_constructor(
            formats.UrlTopic, "query_id", "text", "url"
        ),
        _irds.nfcorpus.NfCorpusQuery: tuple_constructor(
            formats.NFCorpusTopic, "query_id", "title", "all"
        ),
        TrecQuery: tuple_constructor(
            formats.TrecTopic, "query_id", "title", "description", "narrative"
        ),
        _irds.beir.BeirToucheQuery: tuple_constructor(
            formats.TrecTopic, "query_id", "text", "description", "narrative"
        ),
        _irds.beir.BeirSciQuery: tuple_constructor(
            formats.SciDocsTopic,
            "query_id",
            "text",
            "authors",
            "year",
            "cited_by",
            "references",
        ),
        _irds.tweets2013_ia.TrecMb13Query: tuple_constructor(
            formats.TrecMb13Query, "query_id", "query", "time", "tweet_time"
        ),
        _irds.tweets2013_ia.TrecMb14Query: tuple_constructor(
            formats.TrecMb14Query,
            "query_id",
            "query",
            "time",
            "tweet_time",
            "description",
        ),
        _irds.dpr_w100.DprW100Query: tuple_constructor(
            formats.DprW100Query, "query_id", "text", "answers"
        ),
    }

    HANDLERS = {
        cls: partial(SimpleTopicsHandler, converter)
        for cls, converter in CONVERTERS.items()
    }

    def count(self):
        return self.dataset.queries_count()

    @cached_property
    def topic_recordtype(self) -> RecordType:
        return record_type(IDItem, self.handler.target_cls)

    @cached_property
    def handler(self):
        handler = Topics.HANDLERS[self.dataset.queries_cls()](self)
        return handler

    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        return self.handler.topic_int(internal_topic_id)

    def topic_ext(self, external_topic_id: str) -> TopicRecord:
        """Returns a document given its external ID"""
        return self.handler.topic_ext(external_topic_id)

    def iter(self) -> Iterator[TopicRecord]:
        """Returns an iterator over topics"""
        return self.handler.iter()


class TrecBackgroundLinkingTopicsHandler(TopicsHandler):
    def __init__(self, dataset):
        self.dataset = dataset

    @cached_property
    def ext2records(self):
        return {record[IDItem].id: record for record in self.records}

    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        return self.records[internal_topic_id]

    def topic_ext(self, external_topic_id: str) -> TopicRecord:
        """Returns a document given its external ID"""
        return self.ext2records[external_topic_id]

    def iter(self) -> Iterator[ir.TopicRecord]:
        """Returns an iterator over topics"""
        return iter(self.records)

    @cached_property
    def records(self):
        try:
            records = []

            for query in self.dataset.dataset.queries_iter():
                topic = Record(
                    IDItem(query.query_id),
                    # Following BEIR documentation, we use title of documents as queries: https://github.com/beir-cellar/beir/blob/main/examples/dataset/README.md#queries-and-qrels
                    SimpleTextItem(
                        self.dataset.dataset.docs_store().get(query.doc_id).title
                    ),
                    UrlItem(query.url),
                )
                records.append(topic)
        except Exception:
            logging.exception("Error while computing topic records")
            raise

        return records


Topics.HANDLERS.update(
    {_irds.wapo.TrecBackgroundLinkingQuery: TrecBackgroundLinkingTopicsHandler}
)


class CastTopicsHandler(TopicsHandler):
    def __init__(self, dataset):
        self.dataset = dataset

    @cached_property
    def ext2records(self):
        return {record[IDItem].id: record for record in self.records}

    def topic_int(self, internal_topic_id: int) -> TopicRecord:
        """Returns a document given its internal ID"""
        return self.records[internal_topic_id]

    def topic_ext(self, external_topic_id: str) -> TopicRecord:
        """Returns a document given its external ID"""
        return self.ext2records[external_topic_id]

    def iter(self) -> Iterator[ir.TopicRecord]:
        """Returns an iterator over topics"""
        return iter(self.records)

    @cached_property
    def records(self):
        try:
            topic_number = None
            node = None
            conversation = []
            records = []

            for query in self.dataset.dataset.queries_iter():
                decontextualized = DecontextualizedDictItem(
                    "manual",
                    {
                        "manual": query.manual_rewritten_utterance,
                        "auto": query.automatic_rewritten_utterance,
                    },
                )

                is_new_conversation = topic_number != query.topic_number

                topic = Record(
                    IDItem(query.query_id),
                    SimpleTextItem(query.raw_utterance),
                    decontextualized,
                    ConversationHistoryItem(
                        [] if is_new_conversation else node.conversation(False)
                    ),
                    EntryType.USER_QUERY,
                )

                if is_new_conversation:
                    conversation = []
                    node = ConversationTreeNode(topic)
                    topic_number = query.topic_number
                else:
                    node = node.add(ConversationTreeNode(topic))

                records.append(topic)

                conversation.append(node)
                node = node.add(
                    ConversationTreeNode(
                        Record(
                            AnswerDocumentID(self.get_canonical_result_id(query)),
                            EntryType.SYSTEM_ANSWER,
                        )
                    )
                )
                conversation.append(node)
        except Exception:
            logging.exception("Error while computing topic records")
            raise

        return records

    @staticmethod
    def get_canonical_result_id():
        return None


class Cast2020TopicsHandler(CastTopicsHandler):
    @staticmethod
    def get_canonical_result_id(query: _irds.trec_cast.Cast2020Query):
        return query.manual_canonical_result_id


class Cast2021TopicsHandler(CastTopicsHandler):
    @staticmethod
    def get_canonical_result_id(query: _irds.trec_cast.Cast2021Query):
        return query.canonical_result_id


class Cast2022TopicsHandler(CastTopicsHandler):
    def __init__(self, dataset):
        self.dataset = dataset

    @cached_property
    def records(self):
        try:
            records = []
            nodes: Dict[str, ConversationTreeNode] = {}

            for (
                query
            ) in (
                self.dataset.dataset.queries_iter()
            ):  # type: _irds.trec_cast.Cast2022Query
                parent = nodes[query.parent_id] if query.parent_id else None

                if query.participant == "User":
                    topic = Record(
                        IDItem(query.query_id),
                        SimpleTextItem(query.raw_utterance),
                        DecontextualizedDictItem(
                            "manual",
                            {
                                "manual": query.manual_rewritten_utterance,
                            },
                        ),
                        ConversationHistoryItem(
                            parent.conversation(False) if parent else []
                        ),
                        EntryType.USER_QUERY,
                    )
                    node = ConversationTreeNode(topic)
                    records.append(topic)
                else:
                    node = ConversationTreeNode(
                        Record(
                            AnswerEntry(query.response),
                            EntryType.SYSTEM_ANSWER,
                        )
                    )

                nodes[query.query_id] = node
                if parent:
                    parent.add(node)
        except Exception:
            logging.exception("Error while computing topic records")
            raise

        return records


Topics.HANDLERS.update(
    {
        # _irds.trec_cast.Cast2019Query: Cast2019TopicsHandler,
        _irds.trec_cast.Cast2020Query: Cast2020TopicsHandler,
        _irds.trec_cast.Cast2021Query: Cast2021TopicsHandler,
        _irds.trec_cast.Cast2022Query: Cast2022TopicsHandler,
    }
)


class CastDocHandler:
    def check(self, cls):
        assert issubclass(cls, _irds.trec_cast.CastDoc)

    @cached_property
    def target_cls(self):
        return formats.TitleUrlDocument

    def __call__(self, _, doc: _irds.trec_cast.CastDoc):
        return Record(
            IDItem(doc.doc_id), formats.SimpleTextItem(" ".join(doc.passages))
        )


class CastPassageDocHandler:
    def check(self, cls):
        assert issubclass(cls, _irds.trec_cast.CastPassageDoc)

    @cached_property
    def target_cls(self):
        return formats.TitleUrlDocument

    def __call__(self, _, doc: _irds.trec_cast.CastPassageDoc):
        return Record(
            IDItem(doc.doc_id),
            formats.TitleUrlDocument(doc.text, doc.title, doc.url),
        )


Documents.CONVERTERS[_irds.trec_cast.CastDoc] = CastDocHandler()
Documents.CONVERTERS[_irds.trec_cast.CastPassageDoc] = CastPassageDocHandler()


class Adhoc(ir.Adhoc, IRDSId):
    pass


class AdhocRun(ir.AdhocRun, IRDSId):
    pass


class TrainingTriplets(ir.TrainingTriplets, IRDSId):
    """Training triplets from IR Dataset"""

    CONVERTERS = {
        GenericDocPair: lambda qid, doc1_id, doc2_id: (
            create_record(id=qid),
            create_record(id=doc1_id),
            create_record(id=doc2_id),
        )
    }

    @cached_property
    def topic_recordtype(self) -> RecordType:
        """The set of records for topics"""
        return record_type(IDItem)

    @cached_property
    def document_recordtype(self) -> RecordType:
        """The class for documents"""
        return record_type(IDItem)

    @cached_property
    def converter(self):
        return TrainingTriplets.CONVERTERS[self.dataset.docpairs_cls()]

    def iter(
        self,
    ) -> Iterator[Tuple[ir.TopicRecord, ir.DocumentRecord, ir.DocumentRecord]]:
        ds = self.dataset

        logging.info("Starting to generate triplets")
        yield from (self.converter(*entry) for entry in ds.docpairs_iter())
        logging.info("Ending triplet generation")

    def count(self):
        """Returns the length or None"""
        return self.dataset.docpairs_count()



# See documentation on https://datamaestro.readthedocs.io

import re
import json
from pathlib import Path
from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro.data.ml import Supervised
from datamaestro.download import reference
from datamaestro.download.archive import zipdownloader
from datamaestro.download.wayback import wayback_documents
from datamaestro.utils import HashCheck
from datamaestro_text.data.conversation.qrecc import QReCCDataset
from datamaestro_text.datasets.irds.data import (
    LZ4JSONLDocumentStore,
    SimpleJsonDocument,
)
from datamaestro_text.datasets.irds.helpers import lz4docstore_builder


@datatags("conversation", "context", "query")
@datatasks("query rewriting")
@zipdownloader(
    "data",
    "https://github.com/apple/ml-qrecc/raw/main/dataset/qrecc_data.zip",
    checker=HashCheck("f88fcc7ef3678cd6312080389c8abd67"),
)
@dataset(
    url="https://github.com/apple/ml-qrecc",
    doi="https://doi.org/10.48550/arXiv.2010.04898",
    id="",
)
def main(data: Path) -> Supervised[QReCCDataset, None, QReCCDataset]:
    """Open-Domain Question Answering Goes Conversational via Question Rewriting

    We introduce QReCC (Question Rewriting in Conversational Context), an
    end-to-end open-domain question answering dataset comprising of 14K
    conversations with 81K question-answer pairs. The goal of this dataset is to
    provide a challenging benchmark for end-to-end conversational question
    answering that includes the individual subtasks of question rewriting,
    passage retrieval and reading comprehension
    """
    return Supervised.C(
        train=QReCCDataset.C(path=data / "qrecc_train.json"),
        test=QReCCDataset.C(path=data / "qrecc_test.json"),
    )


@dataset(
    url="https://github.com/apple/ml-qrecc",
    doi="https://doi.org/10.48550/arXiv.2010.04898",
)
class Content(LZ4JSONLDocumentStore):
    """QReCC mentionned URLs content"""
    @staticmethod
    def __create_dataset__(dataset, options=None):
        ds = reference(reference=main).setup(dataset, options)
        documents_path = wayback_documents(
            "20191127", lambda: Content._urls(ds), name="wayback.jsonl"
        ).setup(dataset, options)

        store_path = lz4docstore_builder(
            "store",
            lambda: Content._documents(documents_path),
            SimpleJsonDocument,
            "id",
        ).setup(dataset, options)

        return Content(jsonl_path=store_path)

    @staticmethod
    def _documents(path: Path):
        """Iterates over documents from wayback"""
        with path.open("rt") as fp:
            for line in fp:
                yield SimpleJsonDocument(**json.loads(line))

    @staticmethod
    def _urls(supervised: Supervised[QReCCDataset, None, QReCCDataset]):
        urls = set()
        for ds in [supervised.train, supervised.test]:
            for entry in ds.entries():
                if entry.answer_url:
                    url = re.sub("#.*$", "", entry.answer_url)
                    urls.add(url)
        return urls



from experimaestro import Param
from datamaestro.data import Base, File
import datamaestro.data.csv as csv


class RatedItems(Base):
    ratings: Param[File]


class Movielens(RatedItems):
    links: Param[csv.Generic]
    movies: Param[csv.Generic]
    tags: Param[csv.Generic]



# See documentation on https://datamaestro.readthedocs.io

import bz2
from datamaestro.download import reference
from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro_text.data.conversation.base import ConversationUserTopics
from datamaestro_text.data.ir import Adhoc

from datamaestro.utils import HashCheck
from datamaestro.context import DatafolderPath
from datamaestro.download.single import filedownloader
from datamaestro_text.data.conversation.ikat import IkatConversations
from datamaestro.download.links import linkfolder

from datamaestro_text.data.ir.stores import IKatClueWeb22DocumentStore
from datamaestro_text.data.ir.trec import TrecAdhocAssessments
from datamaestro_text.datasets.irds.helpers import lz4docstore_builder


@dataset(as_prepare=True)
def clueweb22(dataset, options=None) -> IKatClueWeb22DocumentStore:
    # Number of documents in the dataset
    count = 116_838_987

    jsonl_folder = linkfolder(
        "documents", [DatafolderPath("gov.nist.trec.ikat.clueweb22", "jsonl")]
    ).setup(dataset, options)
    store_path = lz4docstore_builder(
        "store",
        IKatClueWeb22DocumentStore.generator(
            jsonl_folder,
            jsonl_folder / "ikat_2023_passages_jsonl.sha256sums",
            jsonl_folder / "ikat_2023_passages_hashes.tsv.bz2",
        ),
        IKatClueWeb22DocumentStore.Document,
        "id",
        count_hint=count,
    ).setup(dataset, options)

    return IKatClueWeb22DocumentStore.C(path=store_path, count=count)


@datatags("conversation", "context", "query")
@datatasks("conversational search", "query rewriting")
@reference("documents", clueweb22)
@filedownloader(
    "topics.json",
    "https://raw.githubusercontent.com/irlabamsterdam/iKAT/refs/heads/main/2025/data/2025_test_topics.json",
    checker=HashCheck("16f8444a8d0a8dfe0090f478f185a63c"),
)
@dataset(
    id="2025",
    url="https://github.com/irlabamsterdam/iKAT/tree/main/2025",
)
def test_2025(topics, documents) -> Adhoc.C:
    """Question-in-context rewriting

    iKAT is a test dataset for question-in-context rewriting that consists of
    questions each given in a dialog context together with a context-independent
    rewriting of the question.
    """
    return Adhoc.C(
        topics=ConversationUserTopics.C(conversations=IkatConversations.C(path=topics)),
        # TODO: add when available
        assessments=TrecAdhocAssessments.C(path="/to/do"),
        documents=documents,
    )


@datatags("conversation", "context", "query")
@datatasks("conversational search", "query rewriting")
@reference("documents", clueweb22)
@filedownloader(
    "qrels",
    "https://trec.nist.gov/data/ikat/2024-qrels.txt",
    checker=HashCheck("57f958903ed1c12bbac207f62800814f"),
)
@filedownloader(
    "topics.json",
    "https://raw.githubusercontent.com/irlabamsterdam/iKAT/refs/heads/main/2024/data/2024_test_topics.json",
    checker=HashCheck("ad45bc6e7add2081d69ea60a0a4d1203"),
)
@dataset(
    Adhoc,
    id="2024",
    url="https://github.com/irlabamsterdam/iKAT/tree/main/2024",
)
def test_2024(topics, qrels, documents) -> Adhoc.C:
    """iKAT 2024 dataset"""
    return Adhoc.C(
        topics=ConversationUserTopics.C(conversations=IkatConversations.C(path=topics)),
        assessments=TrecAdhocAssessments.C(path=qrels),
        documents=documents,
    )


@datatags("conversation", "context", "query")
@datatasks("conversational search", "query rewriting")
@reference("documents", clueweb22)
@filedownloader(
    "qrels",
    "https://trec.nist.gov/data/ikat/2023-qrels.all-turns.txt",
    checker=HashCheck("79dc121bab25b2245e52a53263e5ad1f"),
)
@filedownloader(
    "topics.json",
    "https://raw.githubusercontent.com/irlabamsterdam/iKAT/refs/heads/main/2023/data/2023_test_topics.json",
    checker=HashCheck("684fa0197cdec8c3cfb6a2e586ab83f6"),
)
@dataset(
    Adhoc,
    id="2023",
    url="https://github.com/irlabamsterdam/iKAT/tree/main/2023",
)
def test_2023(topics, qrels, documents) -> Adhoc.C:
    """iKAT 2023 dataset"""
    return Adhoc.C(
        topics=ConversationUserTopics.C(conversations=IkatConversations.C(path=topics)),
        assessments=TrecAdhocAssessments.C(path=qrels),
        documents=documents,
    )



import bz2
from hashlib import md5, sha256
import json
import logging
from pathlib import Path
from typing import List, NamedTuple
from datamaestro_text.utils.files import TQDMFileReader
from experimaestro import Constant
from datamaestro.record import Record
from datamaestro_text.data.ir.base import (
    DocumentRecord,
    IDItem,
    SimpleTextItem,
    UrlItem,
)
from datamaestro_text.datasets.irds.data import LZ4DocumentStore
from datamaestro_text.data.ir.formats import OrConvQADocument
from tqdm import tqdm


class OrConvQADocumentStore(LZ4DocumentStore):
    class NAMED_TUPLE(NamedTuple):
        id: str
        title: str
        body: str
        aid: str
        bid: int

    lookup_field: Constant[str] = "id"
    fields: Constant[List[str]] = list(NAMED_TUPLE._fields)
    index_fields: Constant[List[str]] = ["id"]

    data_cls = NAMED_TUPLE

    def converter(self, data: NAMED_TUPLE) -> Record:
        fields = data._asdict()
        del fields["id"]
        return Record(OrConvQADocument(**fields), IDItem(data.id))


class IKatClueWeb22DocumentStore(LZ4DocumentStore):
    @staticmethod
    def generator(path: Path, checksums_file: Path, passages_hashes: Path):
        """Returns an iterator over iKAT 2022-25 documents

        :param path: The folder containing the files
        """

        def __iter__():
            errors = False

            assert checksums_file.is_file(), f"{checksums_file} does not exist"
            assert passages_hashes.is_file(), f"{passages_hashes} does not exist"

            # Get the list of files
            with checksums_file.open("rt") as fp:
                files = []
                for line in fp:
                    checksum, filename = line.strip().split()
                    files.append((checksum, filename))
                    if not (path / filename).is_file():
                        logging.error("File %s does not exist", path / filename)
                        errors = True

            assert not errors, "Errors detected, stopping"

            # Check the SHA256 sums
            match checksums_file.suffix:
                case ".sha256sums":
                    hasher_factory = sha256
                case _:
                    raise NotImplementedError(
                        f"Cannot handle {checksums_file.suffix} checksum files"
                    )

            for checksum, filename in tqdm(files):
                logging.info("Checking %s", filename)
                hasher = hasher_factory()
                with (path / filename).open("rb") as fp:
                    while data := fp.read(2**20):
                        hasher.update(data)

                file_checksum = hasher.hexdigest()
                assert file_checksum == checksum, (
                    f"Expected {checksum}, " f"got {file_checksum} for {filename}"
                )

            # Get the MD5 hashes of all the passages
            logging.info("Reading the hashes of all passages")
            with TQDMFileReader(passages_hashes, "rt", bz2.open) as fp:
                passage_checksums = {}
                for line in tqdm(fp):
                    doc_id, passage_no, checksum = line.strip().split()
                    passage_checksums[f"{doc_id}:{passage_no}"] = checksum  # noqa: E231

            # Read the files
            logging.info("Starting to read the files")
            for _, filename in tqdm(files):
                with TQDMFileReader(path / filename, "rt", bz2.open) as jsonl_fp:
                    for line in jsonl_fp:
                        data = json.loads(line)
                        expected = passage_checksums[data["id"]]
                        computed = md5(data["contents"].encode("utf-8")).hexdigest()
                        assert expected == computed, (
                            f"Expected {expected}, "
                            f"got {computed} for passage {data['id']} in {filename}"
                        )
                        yield IKatClueWeb22DocumentStore.Document(**data)

        return __iter__

    class Document(NamedTuple):
        id: str
        contents: str
        url: str

    data_cls = Document
    lookup_field: Constant[str] = "id"
    index_fields: Constant[List[str]] = ["id"]

    def converter(self, data):
        return DocumentRecord(
            IDItem(data.id), SimpleTextItem(data.contents), UrlItem(data.url)
        )



# See documentation on https://datamaestro.readthedocs.io

import gzip
import json
from pathlib import Path
from typing import Iterator
from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro.download.single import filedownloader
from datamaestro.utils import HashCheck


from datamaestro_text.data.conversation.orconvqa import OrConvQADataset
from datamaestro.data.ml import Supervised

from datamaestro_text.data.ir.stores import OrConvQADocumentStore
from datamaestro_text.datasets.irds.helpers import lz4docstore_downloader


@datatags("conversation", "context", "query")
@datatasks("query rewriting")
@filedownloader(
    "train.jsonl",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/preprocessed/train.txt",
    checker=HashCheck("7513a9ef12d8b7a4471166dc4fef77b7"),
)
@filedownloader(
    "dev.jsonl",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/preprocessed/dev.txt",
    checker=HashCheck("7765658995cc9ffd5eb39a400d814b20"),
)
@filedownloader(
    "test.jsonl",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/preprocessed/test.txt",
    checker=HashCheck("0cf3a755f06297b9c02e7db45f8dc8be"),
)
@dataset(
    Supervised,
    url="https://github.com/prdwb/orconvqa-release",
)
def preprocessed(train, dev, test):
    """Open-Retrieval Conversational Question Answering datasets

    OrConvQA is an aggregation of three existing datasets:

    1. the QuAC dataset that offers information-seeking conversations,
    1. the CANARD dataset that consists of context-independent rewrites of QuAC questions, and
    3. the Wikipedia corpus that serves as the knowledge source of answering questions.

    Each dataset is an instance of :class:`datamaestro_text.data.conversation.OrConvQADataset`
    """
    return {
        "train": OrConvQADataset.C(path=train),
        "validation": OrConvQADataset.C(path=dev),
        "test": OrConvQADataset.C(path=test),
    }


def orConvQADocumentReader(source: Path) -> Iterator[OrConvQADocumentStore.NAMED_TUPLE]:
    with gzip.open(source, "rt") as fp:
        for line in fp:
            data = json.loads(line)
            data["body"] = data.pop("text")
            yield OrConvQADocumentStore.NAMED_TUPLE(**data)


@lz4docstore_downloader(
    "all_blocks",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/all_blocks.txt.gz",
    orConvQADocumentReader,
    OrConvQADocumentStore.NAMED_TUPLE,
    "id",
    checker=HashCheck("1095a3408690e7bbe4d8a87a2bae6356"),
    size=5_086_902_800,
    count_hint=11_377_951,
)
@dataset(
    OrConvQADocumentStore,
    url="https://github.com/prdwb/orconvqa-release",
)
def passages(all_blocks):
    """orConvQA wikipedia files

    OrConvQA is an aggregation of three existing datasets:

    1. the QuAC dataset that offers information-seeking conversations,
    1. the CANARD dataset that consists of context-independent rewrites of QuAC questions, and
    3. the Wikipedia corpus that serves as the knowledge source of answering questions.
    """
    return {"path": all_blocks, "count": 11_377_951}



from typing import Iterator, List, Optional
from attr import define
import json
from datamaestro.data import File
from datamaestro.record import Record

from datamaestro_text.data.ir.base import (
    IDItem,
    SimpleTextItem,
)


from .base import (
    AnswerDocumentURL,
    AnswerEntry,
    ConversationTree,
    EntryType,
    SimpleDecontextualizedItem,
    SingleConversationTree,
)
from . import ConversationDataset


@define(kw_only=True)
class QReCCDatasetEntry:
    """A query with past history"""

    conversation_no: int
    """Conversation ID"""

    turn_no: int
    """The turn in the conversation"""

    conversation_source: str
    """Conversation source"""

    question: str
    """The last issued query"""

    rewrite: str
    """Manually rewritten query"""

    context: List[str]
    """The list of queries asked by the user"""

    answer: str
    """The answer"""

    answer_url: str
    """The URL containing the answer"""


class QReCCDataset(ConversationDataset, File):
    def entries(self) -> Iterator[QReCCDatasetEntry]:
        """Iterates over re-written query with their context"""
        with self.path.open("rt") as fp:
            data = json.load(fp)

        data = [
            QReCCDatasetEntry(**{key.lower(): value for key, value in entry.items()})
            for entry in data
        ]
        return iter(data)

    def __iter__(self) -> Iterator[ConversationTree]:
        history: List[Record] = []
        current_id: Optional[str] = None

        for entry in self.entries():
            # Creates a new conversation if needed
            if entry.conversation_no != current_id:
                if current_id is not None:
                    history.reverse()
                    yield SingleConversationTree(current_id, history)

                current_id = entry.conversation_no
                history = []

            # Add to current
            history.append(
                Record(
                    IDItem(f"{entry.conversation_no}#{entry.turn_no}"),
                    SimpleTextItem(entry.question),
                    AnswerDocumentURL(entry.answer_url),
                    SimpleDecontextualizedItem(entry.rewrite),
                    EntryType.USER_QUERY,
                )
            )

            history.append(
                Record(
                    AnswerEntry(entry.answer),
                    EntryType.SYSTEM_ANSWER,
                )
            )

        # Yields the last one
        history.reverse()
        yield SingleConversationTree(current_id, history)



from typing import Optional
from experimaestro import Param
from datamaestro.data import Base, Folder, File
from datamaestro.data.ml import Supervised


class TrainingText(Supervised):
    """ "A dataset used for training with a train and a test"""

    train: Param[Base]
    test: Param[Optional[Base]] = None
    validation: Param[Optional[Base]] = None


class TextFolder(Folder):
    "A folder composed of texts"
    pass


class TextFile(File):
    """A file composed of texts"""

    pass



import os
from tqdm import tqdm
import gzip
from pathlib import Path


def auto_open(path: Path, mode: str):
    if path.suffix == ".gz":
        return gzip.open(path, mode)
    return path.open(mode)


class CountingWrapper:
    """Wrap a file object to count the actual compressed bytes read."""

    def __init__(self, file_obj):
        self.file_obj = file_obj
        self.bytes_read = 0

    def read(self, size=-1):
        data = self.file_obj.read(size)
        self.bytes_read += len(data)
        return data

    def readline(self, size=-1):
        data = self.file_obj.readline(size)
        self.bytes_read += len(data)
        return data

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    def close(self):
        self.file_obj.close()

    def __getattr__(self, attr):
        return getattr(self.file_obj, attr)


class TQDMBytesReader:
    def __init__(self, file_obj, total_size, **tqdm_kwargs):
        self.file_obj = CountingWrapper(file_obj)
        self.tqdm = tqdm(
            total=total_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            **tqdm_kwargs,
        )

    def _update_progress(self):
        delta = self.file_obj.bytes_read - self.tqdm.n
        if delta > 0:
            self.tqdm.update(delta)

    def read(self, size=-1):
        data = self.file_obj.read(size)
        self._update_progress()
        return data

    def readline(self, size=-1):
        line = self.file_obj.readline(size)
        self._update_progress()
        return line

    def readlines(self, hint=-1):
        lines = self.file_obj.readlines(hint)
        self._update_progress()
        return lines

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    def close(self):
        self.tqdm.close()
        self.file_obj.close()

    def __getattr__(self, attr):
        # Delegate any other attribute to the underlying file object
        return getattr(self.file_obj, attr)


class TQDMFileReader:
    def __init__(self, filepath, mode="rt", file_opener=open, **tqdm_kwargs):
        self.filepath = filepath
        self.mode = mode
        self.file_opener = file_opener
        self.tqdm_kwargs = tqdm_kwargs

    def __enter__(self):
        self.file_obj = self.file_opener(self.filepath, self.mode)
        total_size = os.path.getsize(self.filepath)  # this is compressed size
        self.reader = TQDMBytesReader(
            self.file_obj, total_size=total_size, **self.tqdm_kwargs
        )
        return self.reader

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reader.close()



from typing import Iterator, List
from attr import define
import json
from datamaestro.record import Record
from datamaestro.data import File
from datamaestro_text.data.conversation.base import (
    ConversationDataset,
    ConversationTree,
    SingleConversationTree,
    SimpleDecontextualizedItem,
    EntryType,
)
from datamaestro_text.data.ir import IDItem, SimpleTextItem
import logging


@define(kw_only=True)
class CanardConversation:
    """A query with past history"""

    history: List[str]
    """The list of queries asked by the user"""

    query: str
    """The last issued query"""

    rewrite: str
    """Manually rewritten query"""

    dialogue_id: str
    """Conversation identifier"""

    query_no: int
    """Question number"""


class CanardDataset(ConversationDataset, File):
    """A dataset in the CANARD JSON format

    The CANARD dataset is composed of
    """

    def entries(self) -> Iterator[CanardConversation]:
        """Iterates over re-written query with their context"""
        with self.path.open("rt") as fp:
            data = json.load(fp)

        for entry in data:
            yield CanardConversation(
                history=entry["History"],
                query=entry["Question"],
                rewrite=entry["Rewrite"],
                dialogue_id=entry["QuAC_dialog_id"],
                query_no=entry["Question_no"],
            )

    def __iter__(self) -> Iterator[ConversationTree]:
        history: list[Record] = []
        current_id = None

        for entry in self.entries():
            # Check if current conversation, otherwise we are OK
            if current_id != entry.dialogue_id:
                if current_id is not None:
                    history.reverse()
                    yield SingleConversationTree(current_id, history)
                    history = []

                current_id = entry.dialogue_id

            if not history:
                # First round
                # The two first items are the wikipedia title and section,
                # we interpret them as two user queries
                assert len(entry.history) == 2
                history.extend(
                    Record(
                        SimpleTextItem(text),
                        EntryType.USER_QUERY,
                    )
                    for text in entry.history
                )
            else:
                # The utterance before the last is the last user query
                assert (
                    entry.history[-2] == history[-1][SimpleTextItem].text
                ), f"{entry.dialogue_id} {entry.history} / {history[-4:-1]}"

                # The last utterance is the system side
                history.append(
                    Record(SimpleTextItem(entry.history[-1]), EntryType.SYSTEM_ANSWER)
                )

            assert len(entry.history) == len(history)

            # Add to current
            history.append(
                Record(
                    IDItem(f"{entry.dialogue_id}-{entry.query_no}"),
                    SimpleTextItem(entry.query),
                    SimpleDecontextualizedItem(entry.rewrite),
                    EntryType.USER_QUERY,
                )
            )

        if current_id:
            yield SingleConversationTree(current_id, history)



from typing import Iterator, List
from attr import define, field
import json
import logging
from datamaestro.data import File
from datamaestro.record import Record

from datamaestro_text.data.ir import Topics
from datamaestro_text.data.ir.base import (
    IDItem,
    SimpleTextItem,
)


from .base import (
    AnswerEntry,
    ConversationTree,
    EntryType,
    SimpleDecontextualizedItem,
    SingleConversationTree,
)
from . import ConversationDataset

# Keys to change in the dataset entries for compatibility across different years

KEY_MAPPINGS = {
    # Keys to replace: Target Key
    "turns": "responses",
    "utterance": "user_utterance",
    "ptkb_provenance": "relevant_ptkbs",
    "response_provenance": "citations",
}


def norm_dict(entry: dict) -> dict:
    """Convert keys in the entry to match the expected format."""
    normalized = {}
    for k, v in entry.items():
        # Check for direct mapping, then try lowercase mapping
        new_key = KEY_MAPPINGS.get(k) or KEY_MAPPINGS.get(k.lower()) or k.lower()
        normalized[new_key] = v
    return normalized


@define(kw_only=True)
class IkatConversationEntry:
    """A query with past history"""

    turn_id: int
    """Turn number in the conversation"""

    user_utterance: str
    """The last issued query"""

    resolved_utterance: str
    """Manually rewritten query"""

    response: str
    """The system response to the query"""

    relevant_ptkbs: List[str]
    """The list of relevant personal knowledge bases for the query"""

    citations: List[str]
    """The list of citations for the response"""


@define(kw_only=True)
class IkatConversationTopic:
    """A query with past history"""

    number: str
    """Conversation ID"""

    title: str
    """Title of the conversation"""

    ptkb: str
    """The personal knowledge base associated with the user"""

    responses: List[IkatConversationEntry] = field(
        converter=lambda items: [
            IkatConversationEntry(**item) if isinstance(item, dict) else item
            for item in map(norm_dict, items)
        ]
    )
    """The list of responses to the query"""


class IkatConversations(ConversationDataset, File):
    """A dataset containing conversations from the IKAT project"""

    """Keys to change in the dataset entries for compatibility across different years"""

    def entries(self) -> Iterator[IkatConversationTopic]:
        """Reads all conversation entries from the dataset file."""
        with self.path.open("rt") as fp:
            raw_data = json.load(fp)

        logging.debug("Loaded %d entries from %s", len(raw_data), self.path)
        logging.debug(f"raw data has keys {raw_data[0].keys()}")

        for entry in raw_data:
            try:
                normalized_entry = norm_dict(entry)
                yield IkatConversationTopic(**normalized_entry)
            except Exception as e:
                logging.warning(f"Failed to parse entry: {e}")
                raise e

    def __iter__(self) -> Iterator[ConversationTree]:
        for entry in self.entries():
            history: List[Record] = []

            for turn in entry.responses:
                turn: IkatConversationEntry = turn  # Ensure type is correct
                query_id = f"{entry.number}_{turn.turn_id}"

                # USER QUERY record
                history.append(
                    Record(
                        IDItem(query_id),
                        SimpleTextItem(turn.user_utterance),
                        SimpleDecontextualizedItem(turn.resolved_utterance),
                        EntryType.USER_QUERY,
                    )
                )

                # Build citation info (stubbed relevance to match format)
                relevances = {}
                if turn.relevant_ptkbs:
                    # Example: just use first as relevant (can be improved)
                    relevances[0] = (0, None)  # No position info in this structure

                # SYSTEM ANSWER record
                history.append(
                    Record(
                        AnswerEntry(turn.response),
                        EntryType.SYSTEM_ANSWER,
                    )
                )

            # Ensure reverse if needed for compatibility (optional)
            history.reverse()
            yield SingleConversationTree(entry.number, history)
